#!/usr/bin/env python

"""
Object set support.

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import operator

try:
    set
except NameError:
    from sets import Set as set

class ObjectSet:

    "A set of objects with optional associated data."

    def __init__(self, d=None):
        if d is None:
            self.objects = {}
        elif hasattr(d, "items"):
            self.objects = dict(d)
        else:
            self.objects = {}
            for key in d:
                self.add(key)

    def __repr__(self):
        out = ["{"]
        first = True
        for key, value in self.items():
            if not first:
                out.append(", ")
            else:
                first = False
            out.append(repr(key))
            if value:
                out.append(" : ")
                out.append(repr(value))
        out.append("}")
        return "".join(out)

    def __iter__(self):
        return iter(self.keys())

    def __nonzero__(self):
        return self.objects != {}

    # List methods.

    def __add__(self, other):
        obj = ObjectSet(self)
        for key in other:
            obj.add(key)
        return obj

    # Set membership and comparisons.

    def __hash__(self):
        l = self.keys()
        l.sort()
        return hash(tuple(l))

    def __eq__(self, other):
        if hasattr(other, "objects"):
            return self.objects == other.objects
        else:
            return set(self.objects.keys()) == set(other)

    # Set methods.

    def add(self, obj):
        if not self.has_key(obj):
            self[obj] = set()

    def issubset(self, other):
        return set(self).issubset(other)

    # Dictionary and related methods.

    def __getitem__(self, key):
        return self.objects[key]

    def get(self, key, default=None):
        return self.objects.get(key, default)

    def __setitem__(self, key, value):
        self.objects[key] = value

    def has_key(self, key):
        return self.objects.has_key(key)

    def keys(self):
        return self.objects.keys()

    def values(self):
        return self.objects.values()

    def items(self):
        return self.objects.items()

    def update(self, other):

        # Combining dictionary-like objects involves combining values.

        if hasattr(other, "items"):
            for key, value in other.items():
                self[key] = add_sets(value, self.get(key, []))

        # Combining sequence-like objects involves just adding members.

        else:
            for key in other:
                self.add(key)

    def merge(self, other):

        """
        Merge this object set with an 'other' set, combining the values where
        possible, and incorporating values present in only one of the sets.
        """

        return combine(self, other, ObjectSet(), add_sets)

def deepen_mapping_dict(d):

    "Convert the values of 'd' to be elements of a potentially larger set."

    new_dict = {}
    for key, value in d.items():
        if value is None:
            new_dict[key] = None
        else:
            new_dict[key] = ObjectSet([value])
    return new_dict

def merge_mapping_dicts(dicts):

    "Merge the given 'dicts' mapping keys to sets of objects."

    new_dict = {}
    update_mapping_dict(new_dict, dicts)
    return new_dict

def update_mapping_dict(new_dict, dicts):

    """
    Update 'new_dict' with the contents of the set dictionary 'dicts'.
    None entries may be incorporated into object sets. For example:

    d1: {'a' : None}
    d2: {'a' : x}
    ->  {'a' : {None, x}}

    Here, None might be used to represent an empty set and x may be an existing
    set.
    """

    for old_dict in dicts:
        for key, value in old_dict.items():

            # Add existing mappings within an object set.

            if not new_dict.has_key(key):
                if value is not None:
                    new_dict[key] = ObjectSet(value)
                else:
                    new_dict[key] = None
            elif new_dict[key] is not None:
                if value is not None:
                    new_dict[key].update(value)
                else:
                    new_dict[key].add(value)
            else:
                if value is not None:
                    objset = new_dict[key] = ObjectSet(value)
                    objset.add(None)

def combine_mapping_dicts(d1, d2):

    """
    Combine dictionaries 'd1' and 'd2' in a resulting dictionary, with the
    values of the contributing dictionaries being themselves combined such that
    a "product" of the values for a given key are stored in the combined
    dictionary.

    For example:

    d1: {'a' : [{'f', 'g'}, {'f', 'h'}], ...}
    d2: {'a' : [{'f'}, {'e', 'f', 'g'}], ...}
    ->  {'a' : [{'f', 'g'}, {'f', 'h'}, {'e', 'f', 'g'}, {'e', 'f', 'g', 'h'}], ...}

    Note that items of 'd2' whose keys are not in 'd1' are not added to 'd1'
    since this, in the context of propagating attribute usage observations,
    would result in spurious usage details being made available in places where
    the names may not have been defined.
    """

    return combine(d1, d2, {}, combine_object_set_lists, True)

def combine(d1, d2, combined, combine_op, only_d1_keys=False):

    """
    Combine dictionaries 'd1' and 'd2' in the 'combined' object provided, using
    the 'combine_op' to merge values from the dictionaries.

    If 'only_d1_keys' is set to a true value, items from 'd2' employing keys not
    in 'd1' will not be added to 'd1'.
    """

    if d2 is not None:
        d2_keys = d2.keys()

        for key in d2_keys:
            if not d1.has_key(key):
                if not only_d1_keys:
                    combined[key] = d2[key]
            else:
                combined[key] = combine_op(d1[key], d2[key])
    else:
        d2_keys = ()

    for key in d1.keys():
        if key not in d2_keys:
            combined[key] = d1[key]

    return combined

def combine_object_set_lists(l1, l2):

    """
    Combine lists of object sets 'l1' and 'l2' to make a product of their
    members.
    """

    # If either list is undefined (indicated by None), return the defined list,
    # or return None if neither is defined.

    if l1 is None:
        if l2 is None:
            return None
        else:
            return l2 + []
    elif l2 is None:
        return l1 + []

    combined = ObjectSet()
    for i1 in l1:
        for i2 in l2:
            combined.add(i1.merge(i2))
    return combined

def add_sets(s1, s2):
    return set(list(s1) + list(s2))

# vim: tabstop=4 expandtab shiftwidth=4
